#include <stdio.h>
#include <time.h>
#include <process.h>
#include <nonansi.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "PsxIO.h"
#include "comm.h"
#include "bigint.h"
#include "RXlib.h"

/*
 * Kanaele
 */

#define RING_LO0	4		/* links  aus Ring raus */
#define RING_LI0	5		/* links  in  Ring rein */
#define RING_LOI	7		/* rechts in  Ring rein */
#define RING_LII	6		/* rechts aus Ring raus */
#define PROC_IN(p)	(8+2*(p))	/* vom Slave p          */
#define PROC_OUT(p)	(9+2*(p))	/* zum Slave p          */

/*
 * #define COMM_IN	4
 * #define COMM_OUT	5
 */

typedef struct {
    int status;
    int aufgabe;
    int bucket;
    int shift;
    long anfang;
    long ende;
    long total;
    long mults;
    clock_t lastsamp;
    clock_t lschange;
    clock_t calctime;
    clock_t waittime;
} procstat;

typedef struct {
    int inuse;
    int numprocs;	/* number of results to collect */
    int aktprocs;	/* number of results already collected */
    bigint value;
} mulbuck;

/*
 * Generelle Verwaltung
 */

int       procs;
int       icount;
procstat *pstat;
int       theend, sofar, step;
comm      pollpac;
clock_t   mtime=0;

/*
 * Zahlen und Buckets
 */

int       numbuckets = 4;
mulbuck  *bucket;
int       domult=0;
int       mult1f=0;
int       mult2f=0;
bigint    mult1;
bigint    mult2;
bigint    rectemp;
bigint    result;

/*
 * RX11-Ausgabe
 */

RXDisplay*dpy;
RXGC      gc;
Window    win;
int       screen;
int       xwidth, xstart, ypp;
int       tpers, xpers, yeffd, tyoff;
clock_t   txmin, tlasts, eflt, tpdash;

/*
 * Defaultwerte
 */

#define TPERS		20			/* ticks per sample      */
#define XPERS		5			/* xwidth per sample     */
#define YPP		((procs>=10)?40:50)	/* y width per processor */
#define XWIDTH		600			/* total x width         */
#define XSTART		80			/* x start               */
#define YEFFD		100			/* size of efficiency    */
#define TYOFF		20			/* Time y offset/size    */
#define TPDASH		500			/* time line             */

#define RXPROCMIN(p)	(ypp*(p)+3*ypp/4)
#define RXPROCMAX(p)	(ypp*(p)+1*ypp/4)
#define RXPROCTXT(p)	(ypp*(p)+2*ypp/4)
#define RXEFFY(i)	(ypp*procs+yeffd/10+8*i*yeffd/(procs*10))
#define RXTIMPOS	(ypp*procs+yeffd+3*tyoff/4)

/*
 * Nuetzliche Macros
 */

#define AKTTIME(s)	((double)(clock()-(s))/(double)CLK_TCK)

#define SEND(o,x,l,c) { \
			int xyz; \
			xyz = PsxWrite (o, (char *) (x), l); \
			if (xyz != l) { \
			  fprintf (stderr, "%8.2lf Fatal error master(%d): %d supposed %d errno=%d\n", \
				   AKTTIME(mtime), __LINE__, xyz, l, errno); \
			  c; \
			} \
		  };

#define RECV(i,x,l,c) { \
		        int xyz; \
			xyz = PsxRead (i, (char *) (x), l); \
			if (xyz != l) { \
			  fprintf (stderr, "%8.2lf Fatal error master(%d): %d supposed %d errno=%d\n", \
				   AKTTIME(mtime), __LINE__, xyz, l, errno); \
			  c; \
			} \
		  };


/*
 * Zeichne die X-Ausgaben
 */

void XMoveLeft (void)
{
    RXGCValues gval;
    RXGC ngc;
    int csamp;

    /*
     * Samples on the move
     */
    csamp = ((xwidth/2)/xpers);
    
    RXCopyArea (dpy, win, win, gc,
		xstart + xwidth - csamp * xpers, 0,
		csamp * xpers, procs*ypp + yeffd + tyoff,
		xstart, 0);
    
    gval.foreground = RXWhitePixel (dpy, screen);
    ngc = RXCreateGC (dpy, win, GCForeground, &gval);
    
    RXFillRectangle (dpy, win, ngc,
		     xstart + xwidth - csamp * xpers, 0,
		     csamp * xpers, procs*ypp + yeffd + tyoff);
    
    RXFreeGC (dpy, ngc);
    
    txmin += csamp * tpers;
}

void XDoSample (void)
{
    clock_t now = clock();
    int index, state;
    char tmpstr[42];

    if (dpy == NULL || now-tlasts < tpers)
      return;

    if (((now-txmin)/tpers) * xpers >= xwidth) {
	XMoveLeft ();
    }

    for (index=0; index<procs; index++) {
	state = pstat[index].status;

	RXDrawLine (dpy, win, gc,
		    xstart + ((pstat[index].lastsamp-txmin)*xpers)/tpers,
		    ((state==IDLE) ? RXPROCMIN(index) : RXPROCMAX(index)),
		    xstart + ((now-txmin)*xpers)/tpers,
		    ((state==IDLE) ? RXPROCMIN(index) : RXPROCMAX(index)));

	pstat[index].lastsamp = now;
    }

    RXDrawLine (dpy, win, gc,
		xstart + ((eflt-txmin)*xpers)/tpers, RXEFFY(icount),
		xstart + ((now -txmin)*xpers)/tpers, RXEFFY(icount));

    if (tlasts/tpdash != now/tpdash) {
	RXDrawLine (dpy, win, gc,
		    xstart+(((now/tpdash)*tpdash-txmin)*xpers)/tpers, 
		    0,
		    xstart+(((now/tpdash)*tpdash-txmin)*xpers)/tpers, 
		    procs*ypp + yeffd);

	sprintf (tmpstr, "%-5.0lf",
		 (double)((now/tpdash)*tpdash-mtime)/(double)CLK_TCK);
	RXDrawString(dpy, win, gc,
		     xstart+(((now/tpdash)*tpdash-txmin)*xpers)/tpers,
		     RXTIMPOS,
		     tmpstr, strlen (tmpstr));
    }

    eflt   = now;
    tlasts = now;
}

void XChangeState (int index, int olds, int news, int oldi, int newi)
{
    clock_t now = clock();

    if (!dpy)
      return;

    if (((now-txmin)/tpers) * xpers >= xwidth) {
	XMoveLeft ();
    }

    news = news;

    /*
     * Setze neuen Prozessorstatus
     */

    RXDrawLine (dpy, win, gc,
		xstart + ((pstat[index].lastsamp-txmin)*xpers)/tpers,
		((olds==IDLE) ? RXPROCMIN(index) : RXPROCMAX(index)),
		xstart + ((now-txmin)*xpers)/tpers,
		((olds==IDLE) ? RXPROCMIN(index) : RXPROCMAX(index)));

    RXDrawLine (dpy, win, gc,
		xstart + ((now-txmin)*xpers)/tpers, RXPROCMIN(index),
		xstart + ((now-txmin)*xpers)/tpers, RXPROCMAX(index));

    pstat[index].lastsamp = now;

    /*
     * Setze neuen Effizienzwert
     */

    RXDrawLine (dpy, win, gc,
		xstart + ((eflt - txmin)*xpers)/tpers, RXEFFY(oldi),
		xstart + ((now  - txmin)*xpers)/tpers, RXEFFY(oldi));

    RXDrawLine (dpy, win, gc,
		xstart + ((now  - txmin)*xpers)/tpers, RXEFFY(oldi),
		xstart + ((now  - txmin)*xpers)/tpers, RXEFFY(newi));

    eflt = now;
}

/*
 * Verschicke ein Multiplikationsintervall
 */

int sendaufgabe (int prozessor, long anfang, long ende)
{
    clock_t now = clock();
    comm mycomm;

    fprintf (stderr, "%8.2lf Calculating   %6ld to %6ld  on  %02d\n",
	     AKTTIME(mtime), anfang, ende, prozessor);

    mycomm.prozessor           = prozessor;
    mycomm.command             = AUFGABE;
    mycomm.aufgabe.anfang      = anfang;
    mycomm.aufgabe.ende        = ende;

    pstat[prozessor].status    = BUSY;
    pstat[prozessor].aufgabe   = AUFGABE;
    pstat[prozessor].bucket    = -1;
    pstat[prozessor].shift     =  0;
    pstat[prozessor].anfang    = mycomm.aufgabe.anfang;
    pstat[prozessor].ende      = mycomm.aufgabe.ende;
    pstat[prozessor].waittime += now - pstat[prozessor].lschange;
    pstat[prozessor].lschange  = now;

    SEND (PROC_OUT(prozessor), &mycomm, sizeof (comm), return 1);

    XChangeState (prozessor, IDLE, BUSY, icount, icount-1);
    icount--;

    return 0;
}

/*
 * Versende eine Multiplikationsaufgabe an einen Prozessor
 */

int sendmultcom (int prozessor, bigint *z1, bigint *z2)
{
    clock_t now = clock();
    comm mycomm;

    fprintf (stderr, "%8.2lf Sending multiply  %6dx%-6d to  %02d\n",
    	     AKTTIME(mtime), z1->length, z2->length, prozessor);

    mycomm.prozessor           = prozessor;
    mycomm.command             = MULT;
    mycomm.multiply.length1    = z1->length - z1->first + 1;
    mycomm.multiply.length2    = z2->length - z2->first + 1;
    mycomm.multiply.first1     = 0;
    mycomm.multiply.first2     = 0;

    pstat[prozessor].status    = BUSY;
    pstat[prozessor].aufgabe   = MULT;
    pstat[prozessor].bucket    = -1;
    pstat[prozessor].shift     = z1->first + z2->first;
    pstat[prozessor].waittime += now - pstat[prozessor].lschange;
    pstat[prozessor].lschange  = now;

    SEND (PROC_OUT(prozessor), &mycomm, sizeof (comm), return 1);
    SEND (PROC_OUT(prozessor),
	  z1->digits + z1->first,
	  mycomm.multiply.length1 * sizeof (short),
	  return 1);
    SEND (PROC_OUT(prozessor),
	  z2->digits + z2->first,
	  mycomm.multiply.length2 * sizeof (short),
	  return 1);

    XChangeState (prozessor, IDLE, BUSY, icount, icount-1);
    icount--;

    return 0;
}

/*
 * Verschicke verteilte Multiplikation
 */

int distmult (int mb, int num, bigint *z1, bigint *z2)
{
    clock_t now = clock ();
    int index, theproc;
    long length, iter;
    comm mycomm;

    fprintf (stderr, "%8.2lf Sending multiply  %6dx%-6d to Bucket %02d [",
	     AKTTIME(mtime), z1->length, z2->length, mb);

    bucket[mb].inuse    = 1;
    bucket[mb].numprocs = num;
    bucket[mb].aktprocs = 0;
    
    for (theproc=0, index=0, iter=z1->first; index<num; index++) {
	while (pstat[theproc].status != IDLE)
	  theproc++;
	
	fprintf (stderr, "%02d%s",
		 theproc,
		 (index==num-1)?"]\n":",");
	
	length = ((z1->length - iter) / (num - index)) + 1;
	
	pstat[theproc].status    = BUSY;
	pstat[theproc].aufgabe   = MULT;
	pstat[theproc].bucket    = mb;
	pstat[theproc].shift     = z2->first + iter;
	pstat[theproc].waittime += now - pstat[theproc].lschange;
	pstat[theproc].lschange  = now;
	
	mycomm.prozessor        = theproc;
	mycomm.command          = MULT;
	mycomm.multiply.length1 = length;
	mycomm.multiply.length2 = z2->length - z2->first + 1;
	mycomm.multiply.first1  = 0;
	mycomm.multiply.first2  = 0;
	
	SEND (PROC_OUT(mycomm.prozessor), &mycomm, sizeof (comm), return 1);
	SEND (PROC_OUT(mycomm.prozessor),
	      z1->digits + iter,
	      mycomm.multiply.length1 * sizeof (short),
	      return 1);
	SEND (PROC_OUT(mycomm.prozessor),
	      z2->digits + z2->first,
	      mycomm.multiply.length2 * sizeof (short),
	      return 1);
	
	XChangeState (mycomm.prozessor, IDLE, BUSY, icount, icount-1);
	icount--;

	iter += length;
	theproc++;
    }
    
    /*
     * Hope this was it
     */
    return 0;
}

/*
 * Entscheide, ob eine Multiplikation auf einem oder auf mehreren
 * Prozessen berechnet werden soll.
 */

int seldistmult (int oneproc, bigint *z1, bigint *z2)
{
  int is1longer, thebuck, longest;
  /*
   * multiply on one processor if there is no free bucket (shouldn't happen),
   * not enough idle processors, or if the task is too tiny to distribute
   */

  for (thebuck=0; thebuck<numbuckets; thebuck++) {
      if (bucket[thebuck].inuse == 0)
	break;
  }

  is1longer = (z1->length-z1->first > z2->length-z2->first) ? 1 : 0;
  longest   = is1longer ? (z1->length-z1->first) : (z2->length - z2->first);

  if (thebuck==numbuckets || icount < 2 || longest < 100)
    return sendmultcom (oneproc, z1, z2);

  return distmult (thebuck, 
		   ((longest/50) > icount) ? icount : (longest/50),
		   is1longer ? z1 : z2,
		   is1longer ? z2 : z1);
}

/*
 * Suche eine schoene Aufgabe aus und verschicke sie
 */

int beschaeftige (void)
{
    long anfang, ende;
    int fidle;

    /*
     * Search idle processor
     */

    for (fidle=0; fidle<procs; fidle++)
      if (pstat[fidle].status == IDLE)
	break;

    if (fidle == procs)
      return 1;

    /*
     * If we have something to multiply, do it.
     */
    if (domult) {
	if (seldistmult (fidle, &mult1, &mult2) != 0) {
	    fprintf (stderr, "Fatal error %d.\n", __LINE__);
	    return 1;
	}
	
	domult = 0;
	mult1f = 0;
	mult2f = 0;
	return 0;
    }

    /*
     * If there's still work to be done, send it.
     */
    if (sofar < theend) {
	anfang = sofar + 1;
	ende   = (sofar + step > theend) ? theend : sofar + step;

	if (sendaufgabe (fidle, anfang, ende) != 0)
	  return 1;

	sofar  = ende;
	return 0;
    }

    /*
     * Nothing left to do
     */
    return -1;
}

/*
 * Hauptschleife. Hier passiert einfach alles.
 */

int calcloop (void)
{
    int res, thebuck;
    clock_t ctime;
    comm mycomm;

    while (sofar < theend || icount != procs) {
	RECV (RING_LII, &mycomm, sizeof (comm), return 1);

	XDoSample ();

	switch (mycomm.command) {
	case POLL:
	    SEND (RING_LI0, &mycomm, sizeof (comm), return 1);
	    break;
	case RESULT: /* result is coming in */
	    ctime   = clock ();
	    thebuck = pstat[mycomm.prozessor].bucket;

	    if (thebuck != -1) {
		/*
		 * Ergebnis einer parallelen Multiplikation
		 */
		if (bucket[thebuck].aktprocs == 0) {
		    /*
		     * Das erste Ergebnis in diesem Eimer, d.h. wir
		     * muessen ihn erst initialisieren.
		     */

		    bi_null (&bucket[thebuck].value);
		    RECV (PROC_IN(mycomm.prozessor),
			  bucket[thebuck].value.digits + mycomm.ergebnis.first,
			  mycomm.ergebnis.length * sizeof (unsigned short),
			  return 1);
		    bucket[thebuck].value.first  = mycomm.ergebnis.first;
		    bucket[thebuck].value.length = 
		      mycomm.ergebnis.length + mycomm.ergebnis.first - 1;
		    bucket[thebuck].aktprocs    += 1;

		    if (pstat[mycomm.prozessor].shift > 0)
		      bi_shift_left (&bucket[thebuck].value,
				     pstat[mycomm.prozessor].shift * 4L);
		}
		else {
		    /*
		     * Im Eimer ist bereits etwas, deswegen muessen wir
		     * den neuen Wert draufaddieren.
		     */
		    bi_null (&rectemp);
		    RECV (PROC_IN(mycomm.prozessor),
			  rectemp.digits + mycomm.ergebnis.first,
			  mycomm.ergebnis.length * sizeof (unsigned short),
			  return 1);
		    rectemp.first  = mycomm.ergebnis.first;
		    rectemp.length = 
		      mycomm.ergebnis.length + mycomm.ergebnis.first - 1;

		    if (pstat[mycomm.prozessor].shift > 0)
		      bi_shift_left (&rectemp,
				     pstat[mycomm.prozessor].shift * 4L);
		    
		    bi_add (&bucket[thebuck].value,
			    &rectemp);
		    
		    bucket[thebuck].aktprocs += 1;
		    /*
		     * Bucket full?
		     */
		    if (bucket[thebuck].aktprocs == bucket[thebuck].numprocs) {
			/*
			 * Der Eimer ist voll, uebernehme ihn als Ergebnis
			 */
			if (mult1f == 0) {
			    bi_assign (&mult1, &bucket[thebuck].value);
			    mult1f = 1;
			}
			else /* if (mult2f == 0) */ {
			    bi_assign (&mult2, &bucket[thebuck].value);
			    mult2f = 1;
			}
			bucket[thebuck].inuse = 0;
		    }
		}
	    }
	    /*
	     * Ergebnis einer Einzelmultiplikation wird reingespielt
	     */
	    else if (mult1f == 0) {
		bi_null (&mult1);
		RECV (PROC_IN(mycomm.prozessor),
		      mult1.digits + mycomm.ergebnis.first, 
		      mycomm.ergebnis.length * sizeof (unsigned short),
		      return 1);
		mult1.length = mycomm.ergebnis.length + mycomm.ergebnis.first - 1;
		mult1.first  = mycomm.ergebnis.first;
		mult1f       = 1;
		
		if (pstat[mycomm.prozessor].shift > 0)
		  bi_shift_left (&mult1,
				 pstat[mycomm.prozessor].shift * 4L);
	    }
	    else /* if (mult2f == 0) */ {
		bi_null (&mult2);
		RECV (PROC_IN(mycomm.prozessor),
		      mult2.digits + mycomm.ergebnis.first,
		      mycomm.ergebnis.length * sizeof (unsigned short),
		      return 1);
		mult2.length = mycomm.ergebnis.length + mycomm.ergebnis.first - 1;
		mult2.first  = mycomm.ergebnis.first;
		mult2f       = 1;
		
		if (pstat[mycomm.prozessor].shift > 0)
		  bi_shift_left (&mult2,
				 pstat[mycomm.prozessor].shift * 4L);
	    }
	    if (mult1f == 1 && mult2f == 1) 
	      domult = 1;
	    
	    /*
	     * give him credit
	     */
	    
	    if (pstat[mycomm.prozessor].aufgabe == AUFGABE) {
		pstat[mycomm.prozessor].total   += mycomm.aufgabe.ende - mycomm.aufgabe.anfang + 1;
		fprintf (stderr, "%8.2lf Got result of %6ld to %6ld from %02d   (%8.2lf secs)\n",
			 AKTTIME(mtime),
			 mycomm.aufgabe.anfang,
			 mycomm.aufgabe.ende,
			 mycomm.prozessor,
			 AKTTIME(pstat[mycomm.prozessor].lschange));
	    }
	    else /* if (pstat[mycomm.prozessor].aufgabe == MULT) */ {
		pstat[mycomm.prozessor].mults ++;
		fprintf (stderr, "%8.2lf Got result of mult    %8d from %02d   (%8.2lf secs)\n",
			 AKTTIME(mtime),
		         mycomm.ergebnis.length - 1,
			 mycomm.prozessor,
			 AKTTIME(pstat[mycomm.prozessor].lschange));
	    }
	    /*
	     * Declare him idle
	     */
	    pstat[mycomm.prozessor].status    = IDLE;
	    pstat[mycomm.prozessor].aufgabe   = 0;
	    pstat[mycomm.prozessor].anfang    = 0;
	    pstat[mycomm.prozessor].ende      = 0;
	    pstat[mycomm.prozessor].bucket    =-1;
	    pstat[mycomm.prozessor].shift     = 0;
	    pstat[mycomm.prozessor].calctime += 
	      ctime - pstat[mycomm.prozessor].lschange;
	    pstat[mycomm.prozessor].lschange  = ctime;

	    XChangeState (mycomm.prozessor, BUSY, IDLE, icount, icount+1);
	    icount++;
	    break;
	} /* switch */

	/*
	 * pass out some work
	 */
	do {
	    res = beschaeftige ();
	}
	while (res == 0 && icount > 0);
    } /* main while */

    /*
     * hey, it looks like we're finished. if we are, the result should
     * be in mult1 or mult2
     */
    if (domult || (mult1f == 1 && mult2f == 1)) {
	fprintf (stderr, "%8.2lf Fatal Error in master. Why are we finished if we aren't?\n", AKTTIME(mtime));
	return 1;
    }
    if (mult1f == 1) {
	bi_assign (&result, &mult1);
	mult1f = 0;
    }
    else if (mult2f == 1) {
	bi_assign (&result, &mult2);
	mult2f = 0;
    }
    else {
	fprintf (stderr, "%8.2lf Fatal Error in master. We're finished but don't have a result.\n", AKTTIME(mtime));
	return 1;
    }

    return 0;
}

int main (void)
{
    clock_t c_begin, c_ende, totc, totw;
    char tmpstr[42];
    comm mycomm;
    int index;
    int res;

    fprintf (stderr, "\nMP-Fakultaet\n\n");

    /*
     * Prozessoren zaehlen
     */
    mycomm.prozessor = 0;
    mycomm.command   = COUNT;

    SEND (RING_LI0, &mycomm, sizeof (comm), return 0);
    RECV (RING_LII, &mycomm, sizeof (comm), return 0);
    fprintf (stderr, "\n%d Prozesse gefunden.\n", procs=mycomm.prozessor);

    if ((pstat = (procstat *) malloc (procs * sizeof (procstat))) == NULL) {
	fprintf (stderr, "ERROR! Couldn't allocate resources.\n");
	bucket = NULL;
	res    = 1;
	goto TheEnd;
    }
    if ((bucket = (mulbuck *) malloc (numbuckets * sizeof (mulbuck))) == NULL) {
	fprintf (stderr, "ERROR! Couldn't allocate resources.\n");
	res    = 1;
	goto TheEnd;
    }

    for (index=0; index<procs; index++) {
	pstat[index].status   = IDLE;
	pstat[index].aufgabe  = 0;
	pstat[index].anfang   = 0;
	pstat[index].bucket   =-1;
	pstat[index].shift    = 0;
	pstat[index].ende     = 0;
	pstat[index].total    = 0;
	pstat[index].mults    = 0;
	pstat[index].calctime = 0;
	pstat[index].waittime = 0;
    }

    for (index=0; index<numbuckets; index++) {
	bi_const (&bucket[index].value);
	bucket[index].inuse   = 0;
    }

    bi_const (&result);
    bi_const (&mult1);
    bi_const (&mult2);
    bi_const (&rectemp);

    if (getenv ("ENDE") && getenv ("STEP")) {
	theend = atol (getenv ("ENDE"));
	step   = atol (getenv ("STEP"));
    }
    else {
	fprintf (stderr, "\n");
	fprintf (stderr, "Bis wohin .....  "); fflush (stdout); fflush (stderr); scanf ("%ld", &theend);
	fprintf (stderr, "Schrittweite ..  "); fflush (stdout); fflush (stderr); scanf ("%ld", &step);
	fprintf (stderr, "\n");
    }
    sofar      = 0;
    domult     = 0;
    mult1f     = 0;
    mult2f     = 0;

    /*
     * Einfuegen des Pollpaketes. Es wird immer nur durch den Ring geschickt
     * und weckt die Prozesse auf, damit sie von Zeit zu Zeit nachschauen
     * koennen, ob ihre Aufgabe beendet ist.
     */
    pollpac.prozessor =  0;
    pollpac.command   = POLL;

    SEND (RING_LI0, &pollpac, sizeof (comm), return 0);

    /*
     * Initialisiere RX11 Ausgabe
     */

    if ((dpy = RXOpenDisplay (NULL)) != NULL) {
	tpers  = getenv ("TPERS") ? 
	  (atoi (getenv ("TPERS")) ? atoi (getenv ("TPERS")) : TPERS) :
	    TPERS;
	xpers  = getenv ("XPERS") ?
	  (atoi (getenv ("XPERS")) ? atoi (getenv ("XPERS")) : XPERS) :
	    XPERS;
	ypp    = getenv ("YPP") ?
	  (atoi (getenv ("YPP")) ? atoi (getenv ("YPP")) : YPP) :
	    YPP;
	xwidth = getenv ("XWIDTH") ?
	  (atoi (getenv ("XWIDTH")) ? atoi (getenv ("XWIDTH")) : XWIDTH) :
	    XWIDTH;
	xstart = getenv ("XSTART") ?
	  (atoi (getenv ("XSTART")) ? atoi (getenv ("XSTART")) : XSTART) :
	    XSTART;
	yeffd  = getenv ("YEFFD") ?
	  (atoi (getenv ("YEFFD")) ? atoi (getenv ("YEFFD")) : YEFFD) :
	    YEFFD;
	tyoff  = getenv ("TYOFF") ?
	  (atoi (getenv ("TYOFF")) ? atoi (getenv ("TYOFF")) : TYOFF) :
	    TYOFF;
	tpdash = getenv ("TPDASH") ?
	  (atoi (getenv ("TPDASH")) ? atoi (getenv ("TPDASH")) : TPDASH) :
	    TPDASH;

	screen = RXDefaultScreen      (dpy);
	gc     = RXDefaultGC          (dpy, screen);
	win    = RXCreateSimpleWindow (dpy,
				       RXRootWindow (dpy, screen),
				       20, 20,
				       xwidth + xstart, 
				       procs*ypp + yeffd + tyoff,
				       5,
				       RXBlackPixel (dpy, screen),
				       RXWhitePixel (dpy, screen));

	RXStoreName (dpy, win, "Transputer Factorial Monitor");

	for (index=0; index<procs; index++) {
	    sprintf (tmpstr, "Proc %02d", index);

	    RXDrawString (dpy, win, gc,
			  10, RXPROCTXT(index),
			  tmpstr, strlen (tmpstr));
	    RXDrawLine   (dpy, win, gc,
			  xstart - 10, RXPROCMIN (index),
			  xstart - 10, RXPROCMAX (index));
	}
	sprintf (tmpstr, "  Eff  ");
	RXDrawString (dpy, win, gc,
		      10, procs*ypp+yeffd/2,
		      tmpstr, strlen (tmpstr));
	sprintf (tmpstr, " Time  ");
	RXDrawString (dpy, win, gc,
		      10, RXTIMPOS,
		      tmpstr, strlen (tmpstr));
	RXDrawLine   (dpy, win, gc,
		      xstart - 10, RXEFFY(0),
		      xstart - 10, RXEFFY(procs));
    }

    icount  = procs;
    c_begin = clock    ();
    mtime   = clock    ();

    txmin = tlasts = eflt = mtime;
    for (index=0; index<procs; index++)
      pstat[index].lastsamp = pstat[index].lschange = mtime;

    res     = calcloop ();
    c_ende  = clock    ();

 TheEnd:
    if (res != 0)
      fprintf (stderr, "Oops! Something went wrong.\n");

    fflush (stdout);
    fflush (stderr);
    sleep  (5);

    fprintf (stderr, "Quitting ...\n");
    /*
     * send quit.
     */
    mycomm.prozessor =  0;
    mycomm.command   = QUIT;

    SEND (RING_LI0, &mycomm, sizeof (comm), return 0);

    for (index=0; index<procs; index++) {
	SEND (PROC_OUT(index), &mycomm, sizeof (comm), return 0);
    }

    do {
	RECV (RING_LII, &mycomm, sizeof (comm), return 0);

	if (mycomm.command != POLL && mycomm.command != QUIT) {
	    fprintf (stderr, "Unexpected packet in master: (%d/%d)\n",
		     mycomm.prozessor, mycomm.command);
	}
    }
    while (mycomm.command != QUIT);

    fflush (stdout);
    fflush (stderr);
    sleep  (5);

    fprintf (stderr, "Quitting finished.\n");

    fprintf (stderr, "\nUtilization statistics:\n");
    fprintf (stderr, " Proc | status | total | calctime | waittime | efficiency | mults \n");
    fprintf (stderr, "------+--------+-------+----------+----------+------------+-------\n");
    for (index=0, totc=0, totw=0; index<procs; index++) {
      totc += pstat[index].calctime;
      totw += pstat[index].waittime;
      fprintf (stderr, "  %2d  |   %2d   | %5ld | %7.2lf  | %7.2lf  |  %8.2lf  |  %3d  \n",
	       index,
	       pstat[index].status,
	       pstat[index].total,
	       ((double)(pstat[index].calctime)/(double)CLK_TCK),
	       ((double)(pstat[index].waittime)/(double)CLK_TCK),
	       ((100.0 * (double) pstat[index].calctime) /
	        ((double) c_ende - c_begin)),
	       pstat[index].mults);
    }
    fprintf (stderr, "\n");
    /*
     *  printf ("Time spent in master calc: %8.5lf secs\n\n",
     *          ((double)mtime/(double)CLK_TCK));
     */
    free   (bucket);
    free   (pstat);

    fflush (stdout);
    fflush (stderr);
    if (res == 0) {
	bi_output (&result, stdout);
	printf ("\nCalculation time:   %8.2lf secs\n",
		((double)(c_ende - c_begin)/(double)CLK_TCK));
	printf ("Spent waiting:      %8.2lf%%\n",
	        (100.0*(double)totw/(double)(c_ende-c_begin)/procs));
	printf ("Spent calculating:  %8.2lf%%\n",
	        (100.0*(double)totc/(double)(c_ende-c_begin)/procs));
    }

    if (dpy) {
	scanf ("%s", tmpstr);
	RXDestroyWindow (dpy, win);
	RXCloseDisplay  (dpy);
    }
}
