#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <process.h>
#include <nonansi.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include "PsxIO.h"
#include "comm.h"
#include "bigint.h"

/*
 * Kanaele
 */

#define RING_LIN	0	/* links  in  den Prozessor rein */
#define RING_LOUT	1	/* links  aus dem Prozessor raus */
#define RING_RIN	4	/* rechts in  den Prozessor rein */
#define RING_ROUT	5	/* rechts aus dem Prozessor raus */
#define MAST_IN		6	/* vom Master                    */
#define MAST_OUT	7	/* zum Master                    */

/*
 * #define COMM_IN	0
 * #define COMM_OUT	1
 */

int    prozessnr;
comm   thisaufgabe;
bigint thisnumber;
bigint mult1, mult2;
int    status = IDLE;
Semaphore SemSync;

#define SEND(o,x,l,c) { \
			int xyz; \
			xyz = PsxWrite (o, (char *) (x), l); \
			if (xyz != l) { \
			  fprintf (stderr, "Fatal error slave %d(%d): %d supposed %d errno=%d\n", \
				   prozessnr, __LINE__, xyz, l, errno); \
			  c; \
			} \
		  };

#define RECV(i,x,l,c) { \
		        int xyz; \
			xyz = PsxRead (i, (char *) (x), l); \
			if (xyz != l) { \
			  fprintf (stderr, "Fatal error slave %d(%d): %d supposed %d errno=%d\n", \
				   prozessnr, __LINE__, xyz, l, errno); \
			  c; \
			} \
		  };

/*
 * Diese Schleife ist fuer die Kommunikation zustaendig. Pakete werden
 * asynchron zum Berechnungsprozess angenommen und weitergeleitet.
 */

void docomm (void)
{
    comm mycomm;

    while (42) {
	RECV (RING_LIN, &mycomm, sizeof (comm), continue);
	SEND (RING_ROUT,&mycomm, sizeof (comm), continue);

	if (mycomm.command == QUIT)
	  break;

	/*
	 * Check if we have finished our work
	 */
	if (status == DONE) {
	    memcpy ((char *) &mycomm, &thisaufgabe, sizeof (comm));
	    mycomm.prozessor  = prozessnr;
	    mycomm.command    = RESULT;
	    status            = IDLE;

	    SEND (RING_ROUT, &mycomm, sizeof (comm), ;);

	    /*
	     * Calc process may continue
	     */
	    Signal (&SemSync);
	}
    }
}

/*
 * Berechnung
 */

void docalc (void)
{
    long iter;

    while (42) {
	RECV (MAST_IN, &thisaufgabe, sizeof (comm), continue);
	
	if (thisaufgabe.command == QUIT)
	  break;

	switch (thisaufgabe.command) {
	case AUFGABE:
	    /*
	     * we have everything we need
	     */
	    break;
	case MULT:
	    bi_null (&mult1);
	    bi_null (&mult2);

	    RECV (MAST_IN,
		  mult1.digits + thisaufgabe.multiply.first1,
		  thisaufgabe.multiply.length1 * sizeof (short),
		  ;);
	    RECV (MAST_IN,
		  mult2.digits + thisaufgabe.multiply.first2,
		  thisaufgabe.multiply.length2 * sizeof (short),
		  ;);
		
	    mult1.length = 
	      thisaufgabe.multiply.length1 + thisaufgabe.multiply.first1 - 1;
	    mult2.length = 
	      thisaufgabe.multiply.length2 + thisaufgabe.multiply.first2 - 1;
	    mult1.first  = thisaufgabe.multiply.first1;
	    mult2.first  = thisaufgabe.multiply.first2;
	    break;
	default:
	    fprintf (stderr, "Illegal command slave %d(%d):%d\n",
		     prozessnr, __LINE__, thisaufgabe.command);
	    break;
	}
	/*
	 * Eine neue Aufgabe erwartet uns!
	 */
	status  = BUSY;
	bi_one (&thisnumber);

	if (thisaufgabe.command == AUFGABE) {
	    for (iter =thisaufgabe.aufgabe.anfang;
		 iter<=thisaufgabe.aufgabe.ende;
		 iter++)
	      bi_mult_long (&thisnumber, iter);
	}
	else if (thisaufgabe.command == MULT) {
	    bi_multiply (&thisnumber, &mult1, &mult2);
	}
	else {
	    fprintf (stderr, "Illegal task in slave %d(%d): %d\n",
		     prozessnr, __LINE__, thisaufgabe.command);
	}
	thisaufgabe.ergebnis.length   = thisnumber.length-thisnumber.first+1;
	thisaufgabe.ergebnis.first    = thisnumber.first;
	
	status = DONE;
	Wait (&SemSync);

	SEND (MAST_OUT,
	      thisnumber.digits + thisnumber.first,
	      thisaufgabe.ergebnis.length * sizeof (short),
	      ;);
    }
}

int main (void)
{
    comm mycomm;

    /*
     * zuerst muessen wir wissen wer wir sind
     */

    RECV (RING_LIN,  &mycomm, sizeof (comm), ;);
    prozessnr = mycomm.prozessor++;
    /* fprintf  (stderr, "Greetings from process %d\n", prozessnr); */
    SEND (RING_ROUT, &mycomm, sizeof (comm), ;);
    
    /*
     * Zahl initialisieren
     */
    bi_const (&thisnumber);
    bi_const (&mult1);
    bi_const (&mult2);

    InitSemaphore (&SemSync, 0);

    /*
     * Rechnerprozess starten
     */

    status = IDLE;
    if (Fork (10000, docalc, 0) == 0)
      fprintf (stderr, "Fork failed for process %d.\n", prozessnr);

    docomm ();
    sleep  (5);

    /* fprintf (stderr, "Goodbye from process %d.\n", prozessnr); */
}
